/*
 * Copyright (c) 1999-2003 WIDE Project
 * All rights reserved.
 *
 * Author : Akimichi OGAWA (akimichi@sfc.wide.ad.jp)
 *          Hitoshi  Irino (   irino@sfc.wide.ad.jp)
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code MUST retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form MUST reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    MUST display the following acknowledgement:
 *      This product includes software developed by Akimichi OGAWA.
 * 4. The name of the author MAY NOT be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
*/

#ifdef  HAVE_CONFIG_H
#include <dvts-config.h>
#endif  /* HAVE_CONFIG_H */

#include <sys/param.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <netinet/in.h> //htonl, htons. ntohl. ntohs

#include <fcntl.h>
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

//#include <dv1394.h>

#include <libraw1394/raw1394.h>
#include <libraw1394/csr.h>

//#include <dvts.h>
#include <ieee1394header.h>

#include "hdvsend_param.h"
#include "flags.h"
#include "rtcp.h"
#include "rtp.h"

#include "ieee1394-struct.h"

#define BUFSIZE 512
#define PACKETSIZE 1024

#ifndef DV1394_IOC_INIT
#define DV1394_IOC_INIT DV1394_INIT
#endif /* DV1394_IOC_INIT */

static int _raw1394_read_loop __P((struct hdvsend_param *));
static int _file_read_loop __P((struct hdvsend_param *));
static int _1394_bus_reset_handler __P((raw1394handle_t, unsigned int));
static enum raw1394_iso_disposition _1394_iso_recv_handler __P((raw1394handle_t,
		                                                unsigned char *,
								unsigned int,
								unsigned char,
								unsigned char,
								unsigned char,
								unsigned int,
								unsigned int));
static enum raw1394_iso_disposition _a_1394_iso_recv_handler __P((raw1394handle_t,
		                                                  unsigned char *,
								  unsigned int,
								  unsigned char,
								  unsigned char,
								  unsigned char,
								  unsigned int,
								  unsigned int));

#if 0
static struct dv1394_init dv1394_init_param;

int
open_ieee1394_dv (struct hdvsend_param *hdvsend_param)
{
  int fd;
  struct dv1394_status dv1394_status;
  
  fd = open(hdvsend_param->ieee1394io.ifname, O_RDONLY);
  if (fd < 0) {
    printf("failed to open device : %s\n", hdvsend_param->ieee1394io.ifname);
    return(-1);
  }

  dv1394_init_param.api_version = DV1394_API_VERSION;
  dv1394_init_param.channel = hdvsend_param->ieee1394io.channel;
  dv1394_init_param.n_frames = 2;
  dv1394_init_param.cip_n = 0;
  dv1394_init_param.cip_d = 0;
  dv1394_init_param.syt_offset = 0;

  if(hdvsend_param->format & DV_FORMAT_PAL){
    dv1394_init_param.format = DV1394_PAL;
  } else {
    dv1394_init_param.format = DV1394_NTSC;
  }
  if (ioctl(fd, DV1394_IOC_INIT, &dv1394_init_param) < 0) {
    perror("ioctl DV1394_IOC_INIT");
    return(-1);
  }
  if (ioctl(fd, DV1394_IOC_START_RECEIVE, NULL) < 0) {
    perror("ioctl DV1394_IOC_START_RECEIVE");
    return(-1);
  }

  if (ioctl(fd, DV1394_IOC_GET_STATUS, &dv1394_status) < 0) {
    perror("ioctl DV1394_IOC_GET_STATUS");
  }

  printf("dv1394 status:\n\tchannel: %d, n_frames: %d, format: %s\n",
	 dv1394_status.init.channel, dv1394_status.init.n_frames, 
	 dv1394_status.init.format == 0?"NTSC":"PAL");
  printf("\tcip_n: %ld, cip_d: %ld, syt_offset %d\n",
	 dv1394_status.init.cip_n, dv1394_status.init.cip_d, 
	 dv1394_status.init.syt_offset);

  hdvsend_param->ieee1394io.fd = fd;

  return(1);
}



int
read_loop_dv_dv1394(struct hdvsend_param *hdvsend_param)
{

  printf("start read loop for DV\n");
  u_int32_t readcount = 0;
  struct timeval tv, tv_prev;

  int n;
  int i;
  int readsize;

  /* frame size of PAL is larger than NTSC */
  u_int32_t dvdata[DV1394_PAL_FRAME_SIZE/4];

  gettimeofday(&tv, NULL);
  gettimeofday(&tv_prev, NULL);

  if (hdvsend_param->format == DV_FORMAT_PAL) {
    readsize = 144000;
  } else if (hdvsend_param->format == DV_FORMAT_NTSC) {
    readsize = 120000;
  } else {
    return(-1);
  }

  while (1) {
    n = read(hdvsend_param->ieee1394io.fd, dvdata, readsize);
    if (n < 1) {
      break;
    }

    if (hdvsend_param->output_fd >= 0){
      n = write(hdvsend_param->output_fd, dvdata, readsize);
      if (n < 1) {
        perror("write");
      }
    }

    for (i=0; i<n; i+=80) {
      proc_dvdif(hdvsend_param, &dvdata[i/4]);
      rtcp_check(hdvsend_param, &readcount, &tv_prev);
    }
  }

  return(-1);
}
#endif 

int
read_loop_dv(struct hdvsend_param *hdvsend_param)
{
  printf("start read loop for DVCR\n");
  return _raw1394_read_loop(hdvsend_param);
}

int
open_ieee1394 (struct hdvsend_param *hdvsend_param)
{
	int numcards;
	printf("libraw1394 version %s\n", raw1394_get_libversion());
	hdvsend_param->ieee1394io.handle = raw1394_new_handle();
	if (hdvsend_param->ieee1394io.handle == NULL) {
		fprintf(stderr,
			"raw1394_new_handle() failed\n");
		return(-1);
	}

	numcards = raw1394_get_port_info(hdvsend_param->ieee1394io.handle,
					 hdvsend_param->ieee1394io.portinfo, MAXPORT);
	if (numcards < 0) {
		printf("raw1394_get_port_info() failed\n");
		return(-1);
	} else if (numcards == 0) {
		fprintf(stderr, "No IEEE1394 cards found\n");
		return(-1);
	}

	if (raw1394_set_port(hdvsend_param->ieee1394io.handle,
				hdvsend_param->ieee1394io.cardnum) < 0) {
		printf("raw1394_set_port() failed\n");
		return(-1);
	}
	printf("using IEEE1394 card number : %d\n", hdvsend_param->ieee1394io.cardnum);
	printf("found %d nodes on IEEE1394 bus\n",
		raw1394_get_nodecount(hdvsend_param->ieee1394io.handle));
	printf("get local node id = %x\n", raw1394_get_local_id(hdvsend_param->ieee1394io.handle));
	printf("get isochronous resource node id = %x\n", raw1394_get_irm_id(hdvsend_param->ieee1394io.handle));

	raw1394_set_userdata(hdvsend_param->ieee1394io.handle, (void *)hdvsend_param);
	raw1394_set_bus_reset_handler(hdvsend_param->ieee1394io.handle,
				      _1394_bus_reset_handler);
	return(1);
}
	
int
read_loop_m2ts(struct hdvsend_param *hdvsend_param)
{
  printf("start read loop for MPEG\n");
  if (strncmp(hdvsend_param->ieee1394io.ifname, "/dev/", 5) == 0)
    return _raw1394_read_loop(hdvsend_param);
  else 
    return _file_read_loop(hdvsend_param);
}

static int
_file_read_loop(struct hdvsend_param *hdvsend_param)
{
  u_int32_t readcount = 0;
  struct timeval tv_prev;
  struct timeval tv;
  struct timeval tv2;
  struct timeval tv_sub;
  u_int32_t _buf[192 / sizeof(u_int32_t ) * 10];
  int n_tsp = 0;
  int n_total = 0;
  int n_cycle = 0;
  int n = 0;
  u_int32_t prev_cycle = 0;

  memset(_buf, 0, sizeof(_buf));
  gettimeofday(&tv_prev, NULL);
  gettimeofday(&tv, NULL);
  while (1) {
    u_int32_t buf[192 / sizeof(u_int32_t)];

    /* RTCP */
    rtcp_check(hdvsend_param, &readcount, &tv_prev);

    n = read(hdvsend_param->ieee1394io.fd, buf, 192);
    if (n == 192) {
      u_int32_t sph = ntohl(buf[0]);
      u_int32_t cycle = ((sph  >> 12) & 0x1fff) % 8000;

      if ((prev_cycle != 0) && (prev_cycle != cycle)) {
	u_int32_t u1;
	u_int32_t u2;

	proc_m2ts_sp(hdvsend_param, _buf, n_tsp * 192);
	n_tsp = 0;
	n_cycle++;
	gettimeofday(&tv2, NULL);
	timersub(&tv2, &tv, &tv_sub);
	
	u1 = 1000000 / 8000 * n_cycle;
	u2 = tv_sub.tv_sec * 1000000 + tv_sub.tv_usec;
	if (u1 > u2) {
	  int32_t udiff = u1 - u2;
	  if (udiff > 15) {
	    usleep(udiff);
	    n_cycle = 0;
	    tv = tv2;
	  }
	}
      }
      memcpy((void *)(_buf + n_tsp * 192 / sizeof(u_int32_t)), (const void *)buf, 192);
      n_tsp++;
      n_total++;
      prev_cycle = cycle;
    } else if (n == 0) {
      printf("EOF\n");
      break;
    } else {
      perror("read");
    }
  }

  return (-1);
}

static int 
_raw1394_read_loop(struct hdvsend_param *hdvsend_param)
{
	u_int32_t readcount = 0;
	struct timeval tv_prev;

	gettimeofday(&tv_prev, NULL);
	raw1394_iso_recv_start(hdvsend_param->ieee1394io.handle, -1, -1, 0);
	while (1) {
		/* RTCP */
		rtcp_check(hdvsend_param, &readcount, &tv_prev);
		raw1394_loop_iterate(hdvsend_param->ieee1394io.handle);
	}
	return(-1);
}

int
detect_1394_format(struct hdvsend_param *hdvsend_param)
{
	int ret;

	if (strncmp(hdvsend_param->ieee1394io.ifname, "/dev/", 5) == 0) {
	  ret = open_ieee1394(hdvsend_param);
	  if (ret < 0) {
	    printf("Failed to prepare IEEE1394 input\n");
	    printf("EXIT\n");
	    exit(EXIT_FAILURE);
	  }

	  ret = raw1394_iso_recv_init(hdvsend_param->ieee1394io.handle,
				      _a_1394_iso_recv_handler, BUFSIZE,
				      PACKETSIZE,
				      hdvsend_param->ieee1394io.channel, RAW1394_DMA_PACKET_PER_BUFFER, -1);
	  if(ret < 0){
	    perror("raw1394_iso_recv_init");
	    exit(EXIT_FAILURE);
	  }
	  
	  raw1394_iso_recv_start(hdvsend_param->ieee1394io.handle, -1, -1, 0);
	  raw1394_loop_iterate(hdvsend_param->ieee1394io.handle);
	  raw1394_iso_stop(hdvsend_param->ieee1394io.handle);
	} else {
	  hdvsend_param->ieee1394io.fd = open (hdvsend_param->ieee1394io.ifname, O_RDONLY, S_IRUSR|S_IWUSR|S_IRGRP|S_IROTH);
	  if (hdvsend_param->ieee1394io.fd < 0) {
	    printf("fail to open %s\n", hdvsend_param->ieee1394io.ifname);
	    exit(EXIT_FAILURE);
	  } else {
	    printf("read from %s\n", hdvsend_param->ieee1394io.ifname);
	    hdvsend_param->ieee1394io.ciphdr.fmt = CIP_FMT_MPEG;
	    hdvsend_param->flags |= SEND_M2TS_SP;
	    return 1;
	  }
	}

	if(hdvsend_param->ieee1394io.ciphdr.fmt == CIP_FMT_DVCR){
		raw1394_iso_shutdown(hdvsend_param->ieee1394io.handle);
		if (hdvsend_param->rtp_pt == 0) hdvsend_param->rtp_pt = DEFAULT_DVCR_RTP_PT;
		/*		ret = open_ieee1394_dv(hdvsend_param);*/
		printf("Format: DVCR\n");
	} else if(hdvsend_param->ieee1394io.ciphdr.fmt == CIP_FMT_MPEG){
		raw1394_iso_shutdown(hdvsend_param->ieee1394io.handle);
		if (hdvsend_param->rtp_pt == 0) hdvsend_param->rtp_pt = DEFAULT_MPEG_RTP_PT;
		printf("Format: MPEG\n");
	} else {
	  printf("Unsupported format %x\n", hdvsend_param->ieee1394io.ciphdr.fmt);
	  printf("EXIT\n");
	  return(-1);
	}

	ret = raw1394_iso_recv_init(hdvsend_param->ieee1394io.handle,
				    _1394_iso_recv_handler, BUFSIZE,
				    PACKETSIZE,
				    hdvsend_param->ieee1394io.channel,
				    RAW1394_DMA_PACKET_PER_BUFFER, -1);

	if (ret < 0) {
	  printf("Failed to prepare IEEE1394 input\n");
	  printf("EXIT\n");
	  return(-1);
	}

	return ret;
}

static enum raw1394_iso_disposition
_1394_iso_recv_handler (raw1394handle_t handle, unsigned char *data,
		        unsigned int len, unsigned char channel,
			unsigned char tag, unsigned char sy,
			unsigned int cycle, unsigned int dropped)
{
  int i;
  struct hdvsend_param *hdvsend_param = (struct hdvsend_param *)raw1394_get_userdata(handle);
  struct ieee1394ciphdr *ciphdr = (struct ieee1394ciphdr *)data;

  if (dropped > 0) printf("dropped: %d\n", dropped);
  
  if (len >= sizeof(struct ieee1394ciphdr)) {
    switch (ciphdr->fmt) {
    case CIP_FMT_MPEG:
	proc_m2ts_sp(hdvsend_param, (u_int32_t *)(data + sizeof(struct ieee1394ciphdr)), len - sizeof(struct ieee1394ciphdr));

	/*	      return RAW1394_ISO_DEFER;  */
	       return RAW1394_ISO_OK; 
      break;
      
    case CIP_FMT_DVCR:
      for (i = 0; i < (len - 8) / 80 ; i++) {
	proc_dvdif(hdvsend_param, (u_int32_t *)(data + 8 + i * 80));
      }
      /* return RAW1394_ISO_DEFER; */
             return RAW1394_ISO_OK; 
      break;
    }
  }
       return RAW1394_ISO_DEFER; 
}

static enum raw1394_iso_disposition
_a_1394_iso_recv_handler (raw1394handle_t handle, unsigned char *data,
		        unsigned int len, unsigned char channel,
			unsigned char tag, unsigned char sy,
			unsigned int cycle, unsigned int dropped)
{
	struct hdvsend_param *hdvsend_param = (struct hdvsend_param *)raw1394_get_userdata(handle);

	if (len > sizeof(struct ieee1394ciphdr)) {
	  memcpy(&hdvsend_param->ieee1394io.ciphdr, data ,sizeof(struct ieee1394ciphdr));
	  return RAW1394_ISO_DEFER;
	}
	return RAW1394_ISO_OK;
}

static int
_1394_bus_reset_handler (raw1394handle_t handle, unsigned int generation)
{
  printf("BUS RESET\n");
  return(0);
}
