/*
 * Copyright (c) 1999-2003 WIDE Project
 * All rights reserved.
 *
 * Author : Akimichi OGAWA (akimichi@sfc.wide.ad.jp)
 *          Hitoshi  IRINO (   irino@sfc.wide.ad.jp)
 *          Koji     OGAWA (  pierre@sfc.wide.ad.jp)
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code MUST retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form MUST reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    MUST display the following acknowledgement:
 *      This product includes software developed by Akimichi OGAWA.
 * 4. The name of the author MAY NOT be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
*/

#ifdef	HAVE_CONFIG_H
#include <dvts-config.h>
#endif	/* HAVE_CONFIG_H */

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>

#include <signal.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <netinet/in.h>

#ifdef	LINUX
#include <getopt.h>
#endif	/* LINUX */

#include "dvts.h"
#include "hdvsend_param.h"
#include "info.h"
#include "sig-behavior.h"
#include "flags.h"
#include "udp.h"
#include "multicast.h"
#include "rtp.h"

#include "ieee1394.h"

#define DEFAULT_DVTS_SERV	"8000"


/* this structure contains all configurations */
struct hdvsend_param hdvsend_param;

int
read_loop(struct hdvsend_param *hdvsend_param)
{
  if (hdvsend_param->ieee1394io.ciphdr.fmt == CIP_FMT_DVCR) {
    return read_loop_dv(hdvsend_param);
  } else if (hdvsend_param->ieee1394io.ciphdr.fmt == CIP_FMT_MPEG) {
    return read_loop_m2ts(hdvsend_param);
  }
  return (-1);
}


int
main(int argc, char *argv[])
{
  /* these are for getopt() */
  extern char *optarg;
  extern int optind;
  int op;

  int i;

  /* containor for the return value of each functions */
  int ret;

  /* audio redundancy value */
  int ar_val;

  char *output_filename = NULL;

  /**********************************************/
  /*            program starts here             */
  /**********************************************/

  /*----------------------*/
  /*     SET DEFAULTS     */
  /*----------------------*/

  memset(&hdvsend_param, 0, sizeof(hdvsend_param));

  /* port number for DV/RTP packets */
  strncpy(hdvsend_param.default_serv, DEFAULT_DVTS_SERV,
	  sizeof(hdvsend_param.default_serv));

  /* TTL (hop limit) for multicast packets */
  hdvsend_param.default_multi_ttl = 1;
  memset(hdvsend_param.default_multi_src_addr, 0,
	 sizeof(hdvsend_param.default_multi_src_addr));

  /* Flags (see dvsend-flags.h for detail) */
  hdvsend_param.flags = USE_RTCP; /* RTCP is enabled by default */

  /* IEEE1394 isochronous channel */
  hdvsend_param.ieee1394io.channel = 0x3f;

  /* IEEE1394 interface name */
  memset(hdvsend_param.ieee1394io.ifname, 0, sizeof(hdvsend_param.ieee1394io.ifname));
  strncpy(hdvsend_param.ieee1394io.ifname, DEFAULTIEEE1394DEV, MAXPATHLEN);

  /* The length of each DV/RTP packet */
  /* The payload length will be (DIFblocks_in_pkt * 80) */
  hdvsend_param.DIFblocks_in_pkt = 17;
  hdvsend_param.mtu = 1500;

  /* The video frame drop rate */
  /* (frame_drop == 1) means full rate, i.e. 100% quality */
  hdvsend_param.frame_drop = 1;

  hdvsend_param.format = DV_FORMAT_NTSC;
  hdvsend_param.output_fd = -1;

  TAILQ_INIT(&hdvsend_param.dest_list);
  TAILQ_INIT(&hdvsend_param.dest_audio_list);

  TAILQ_INIT(&hdvsend_param.rtcp_recv_list);
  TAILQ_INIT(&hdvsend_param.rtcp_recv_audio_list);

  /*---------------------*/
  /* END OF SET DEFAULTS */
  /*---------------------*/

  /* getting options */
  while ((op = getopt(argc, argv, "Avh:64NI:P:s:S:M:m:t:d:D:f:LHD:C:Rr:pi:O:o")) > 0) {
    switch (op) {
    case 'A':
      hdvsend_param.flags |= SEND_M2TS_SP;
        break;
      case 'v':
        show_version();
        exit(1);
        break;

      case 'h':
        add_destination(&hdvsend_param, optarg);
        break;

#ifdef ENABLE_INET6
      case '6':
        hdvsend_param.flags |= USE_IPV6_ONLY;
        break;
#endif /* ENABLE_INET6 */

      case '4':
        hdvsend_param.flags |= USE_IPV4_ONLY;
        break;

      case 'N':
        hdvsend_param.flags |= NO_VIDEO_STREAM;
        break;

      case 'I':
   		memset(hdvsend_param.ieee1394io.ifname, 0, sizeof(hdvsend_param.ieee1394io.ifname));
	    strncpy(&hdvsend_param.ieee1394io.ifname[0], optarg, MAXPATHLEN);
        break;

      case 'P':
        strncpy(hdvsend_param.default_serv, optarg, 
		sizeof(hdvsend_param.default_serv));
        break;
	
    case 'M':
      case 'm':
        strncpy(hdvsend_param.default_multi_src_addr, optarg, 
		sizeof(hdvsend_param.default_multi_src_addr));
        break;

      case 't':
        hdvsend_param.default_multi_ttl = atoi(optarg);
        break;

      case 's':
        hdvsend_param.DIFblocks_in_pkt = atoi(optarg);
        if (hdvsend_param.DIFblocks_in_pkt > 17) {
          printf("Number of DIF blocks in packet too large : %d\n",
                 hdvsend_param.DIFblocks_in_pkt);
          printf("MUST be,  3 < DIFblocks < 18\n");
        }
        else if (hdvsend_param.DIFblocks_in_pkt < 4) {
          printf("Number of DIF blocks in packet too small : %d\n",
                 hdvsend_param.DIFblocks_in_pkt);
          printf("MUST be,  3 < DIFblocks < 18\n");
        }
        break;

      case 'S':
        hdvsend_param.mtu = atoi(optarg);

      case 'd':
        hdvsend_param.flags |= SEPARATED_AUDIO_STREAM;
        strncpy(hdvsend_param.default_audio_serv, optarg, 
		sizeof(hdvsend_param.default_audio_serv));

      case 'D':
        hdvsend_param.flags |= SEPARATED_AUDIO_STREAM;
        add_audio_only_destination(&hdvsend_param, optarg);
        break;

      case 'f':
        hdvsend_param.frame_drop = atoi(optarg);
        if (hdvsend_param.frame_drop < 1) {
          printf("invalid frame drop value : %d\n", hdvsend_param.frame_drop);
          exit(-1);
        }
        break;

      case 'L':
        hdvsend_param.flags |= SHOW_PACKETLOSS;
        break;

      case 'R':
        hdvsend_param.flags &= ~USE_RTCP;
        break;

      case 'r':
        ar_val = atoi(optarg);
        if (ar_val < 0) {
          printf("invalid audio redundancy level : %s\n", optarg);
          exit(1);
        }
        hdvsend_param.audio_redundancy = ar_val;
        break;
      case 'T':
        hdvsend_param.rtp_pt = atoi(optarg);
        if(hdvsend_param.rtp_pt <= 96 || hdvsend_param.rtp_pt >= 127){
	  hdvsend_param.rtp_pt = 0;
        }
        break;
      case 'p':
        hdvsend_param.format = DV_FORMAT_PAL;
        break;
      case 'i': /* interval control */
        hdvsend_param.flags |= USE_RATECONTROL;
        break;

      case 'O':
	hdvsend_param.frame_max = atoi(optarg);
	printf("saving %d DV frames\n", hdvsend_param.frame_max);
	break;

      case 'o':
	output_filename = optarg;
	break;
	
      case 'H':
      default:
        show_usage(argv[0]);
        exit(0);
        break;
    }
  }
  argc -= optind;
  argv += (optind - 1);

  /*----------------------*/
  /*    INIT FUNCTIONS    */
  /*----------------------*/

  for (i=argc; i>0; i--) {
    add_destination(&hdvsend_param, argv[i]);
  }

  /* set signal behaviors */
  set_signal_behavior ();

  /* prepare a UDP socket for DV/RTP */
  if (prepare_udp_socket(&hdvsend_param) < 0) {
    return(-1);
  }

  /* setup multicast parameters */
  /* will do nothing if the destination IP address is not a multicast address */
  multicast_settings(&hdvsend_param);

  if (output_filename == NULL) {
    hdvsend_param.output_fd = fileno(stdout);
  } else {
    hdvsend_param.output_fd = open(output_filename, O_RDWR|O_CREAT, S_IRUSR|S_IWUSR|S_IRGRP|S_IROTH);
  }

  detect_1394_format(&hdvsend_param);
  proc_init(&hdvsend_param);
  read_loop(&hdvsend_param);

  return(ret);
}
