#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <netinet/in.h>

#ifdef LINUX
#include <asm/param.h>
#endif /* LINUX */

#include <rtpvar.h>
#include <dvformat.h>
#include <m2tsformat.h>

#include "hdvsend_param.h"
#include "flags.h"
#include "rtp.h"
#include "udp.h"
#include "dvts.h"

//#define		MAX_PACKET_LENGTH	9216
#define		MAX_PACKET_LENGTH   1500	

/* buffer for UDP */
static u_int32_t outbuf[MAX_PACKET_LENGTH/4];
static int outbuf_len;

/* buffer for UDP (for separate audio) */
static u_int32_t audiobuf[MAX_PACKET_LENGTH/4];
static int audiobuf_len;

static u_int32_t r_outbuf[MAX_PACKET_LENGTH/4];
static int r_outbuf_len;

void	proc_buffer_init __P(());
void	proc_rtp_init __P((struct hdvsend_param *));

void	proc_dvdif_init __P((struct hdvsend_param *));
void	proc_m2ts_init __P((struct hdvsend_param *));

static void _add_dvdif_to_outbuf __P((struct hdvsend_param *, u_int32_t *));
static void _add_dvdif_to_r_outbuf __P((struct hdvsend_param *, u_int32_t *));
static void _add_dvdif_to_audio_outbuf __P((struct hdvsend_param *, u_int32_t *));
static void _dvdif_flush __P((struct hdvsend_param *));

static void _add_m2ts_to_outbuf __P((struct hdvsend_param *, u_int32_t *, size_t));
static void _add_m2ts_sp_to_outbuf __P((struct hdvsend_param *, u_int32_t *, size_t));

static void _dvrtp_sendto_ratecontrol __P((struct hdvsend_param, int));

/*---------------------------------------------*/
/* These functions will process each DIF block */
/*---------------------------------------------*/
static void _process_dv_system_common  __P((struct hdvsend_param *,
                                      u_char, u_char, u_char, u_int32_t *));
static void _process_sct_header  __P((struct hdvsend_param *,
                                      u_char, u_char, u_char, u_int32_t *));
static void _process_sct_subcode __P((struct hdvsend_param *,
                                      u_char, u_char, u_char, u_int32_t *));
static void _process_sct_vaux    __P((struct hdvsend_param *,
                                      u_char, u_char, u_char, u_int32_t *));
static void _process_sct_audio   __P((struct hdvsend_param *,
                                      u_char, u_char, u_char, u_int32_t *));
static void _process_sct_video   __P((struct hdvsend_param *,
                                      u_char, u_char, u_char, u_int32_t *));

/* function container */
static void (*sct_func[5]) __P((struct hdvsend_param *,
                                u_char, u_char, u_char, u_int32_t *));

/* The number of frame        */
/* will be used by frame_drop */
static u_long frame = 1;

/* RTP header */
static rtp_hdr_t *rtphdr;
static rtp_hdr_t *audio_rtphdr;

void
proc_dvdif_init(struct hdvsend_param *hdvsend_param)
{
  sct_func[0] = _process_sct_header;
  sct_func[1] = _process_sct_subcode;
  sct_func[2] = _process_sct_vaux;
  sct_func[3] = _process_sct_audio;
  sct_func[4] = _process_sct_video;
}

void
proc_buffer_init()
{
  srandom(time(0));

  memset(outbuf, 0, sizeof(outbuf));
  outbuf_len = 0;
  rtphdr = (rtp_hdr_t *)outbuf;

  memset(audiobuf, 0, sizeof(audiobuf));
  audiobuf_len = 0;
  audio_rtphdr = (rtp_hdr_t *)audiobuf;

  memset(r_outbuf, 0, sizeof(r_outbuf));
  r_outbuf_len = 0;
}

void
proc_rtp_init(struct hdvsend_param *hdvsend_param)
{
  /* RTP version 2 */
  rtphdr->version = RTP_VERSION;
  audio_rtphdr->version = RTP_VERSION;

  /* padding flag                                              */
  /* Padding will be required if the RTP payload length is NOT */
  /* multiple of 4. There will be no padding in DV/RTP, since  */
  /* DV/RTP payload will always be multiple of 80.             */
  rtphdr->p = 0;
  audio_rtphdr->p = 0;

  rtphdr->pt = hdvsend_param->rtp_pt;
  /*
     There aren't much meaning for doing htons and htonl
     on random values. However, this is done for remebering
     to exchange host byte order to network byte order.
  */

  /* RTP sequence number */
  rtphdr->seq = htons(random());
  audio_rtphdr->seq = htons(random());

  /* RTP timestamp */
  rtphdr->ts  = htonl(random());
  audio_rtphdr->ts  = htonl(random());

  /* RTP SSRC */
  hdvsend_param->rtp_ssrc = random();
  rtphdr->ssrc = htonl(hdvsend_param->rtp_ssrc);
  audio_rtphdr->ssrc = htonl(random());

  memcpy(r_outbuf, outbuf, sizeof(rtp_hdr_t));
}

void 
proc_init(struct hdvsend_param *hdvsend_param)
{
  proc_buffer_init();
  if (hdvsend_param->ieee1394io.ciphdr.fmt == CIP_FMT_DVCR) {
    proc_dvdif_init(hdvsend_param);
  } else if (hdvsend_param->ieee1394io.ciphdr.fmt == CIP_FMT_MPEG) {
    proc_m2ts_init(hdvsend_param);
  }
  proc_rtp_init(hdvsend_param);
}

void
proc_m2ts_init(struct hdvsend_param *hdvsend_param)
{
  /* nothing to do! */
}

void
proc_dvdif(struct hdvsend_param *hdvsend_param, u_int32_t *difblock)
{
  static u_char seq, dseq, dbn, sct;

  if (difblock == NULL) {
    printf("proc_dvdif: difblock == NULL\n");
    return;
  }

  dbn  = (ntohl(difblock[0]) >>  8) & 0xff;
  dseq = (ntohl(difblock[0]) >> 20) & 0xf;
  seq  = (ntohl(difblock[0]) >> 24) & 0xf;
  sct  = (ntohl(difblock[0]) >> 29) & 0x7;

  if (sct > 0x4) {
    printf("WARNING : INVALID DV SCT [%d]\n", sct);
    return;
  }

  (sct_func[sct]) (hdvsend_param, seq, dseq, dbn, difblock);
}

void
proc_m2ts(struct hdvsend_param *hdvsend_param, u_int32_t *m2ts, size_t length)
{

}


void
proc_m2ts_sp(struct hdvsend_param *hdvsend_param, u_int32_t  *m2ts_sp, size_t length)
{
  if (length) {
    if (hdvsend_param->flags & SEND_M2TS_SP)
      _add_m2ts_sp_to_outbuf(hdvsend_param, m2ts_sp, length);
    else
      _add_m2ts_to_outbuf(hdvsend_param, m2ts_sp, length);    
  }
}

static void
_process_dv_system_common(struct hdvsend_param *hdvsend_param, u_char seq,
		          u_char dseq, u_char dbn, u_int32_t *difblock)
{
  if (!(hdvsend_param->frame_drop > 1 &&
      (frame % hdvsend_param->frame_drop) != 0)) {
    _add_dvdif_to_outbuf(hdvsend_param, difblock);
    /*
    if (hdvsend_param->flags & SEPARATED_AUDIO_STREAM) {
      _add_dvdif_to_withoutaudio_audiobuf(hdvsend_param, difblock);
    }
    */
  }
}

/* This will process DV HEADER DIF blocks */
static void
_process_sct_header (struct hdvsend_param *hdvsend_param,
                     u_char seq, u_char dseq, u_char dbn, u_int32_t *difblock)
{
  _process_dv_system_common(hdvsend_param, seq, dseq, dbn, difblock); 
}

/* This will process DV SUBCODE DIF blocks */
static void
_process_sct_subcode (struct hdvsend_param *hdvsend_param,
                      u_char seq, u_char dseq, u_char dbn, u_int32_t *difblock)
{
  _process_dv_system_common(hdvsend_param, seq, dseq, dbn, difblock); 
}

/* This will process DV VAUX DIF blocks */
static void
_process_sct_vaux (struct hdvsend_param *hdvsend_param,
                   u_char seq, u_char dseq, u_char dbn, u_int32_t *difblock)
{
  _process_dv_system_common(hdvsend_param, seq, dseq, dbn, difblock); 
}

/* This will process DV AUIDO DIF blocks */
static void
_process_sct_audio (struct hdvsend_param *hdvsend_param,
                    u_char seq, u_char dseq, u_char dbn, u_int32_t *difblock)
{
  if (!(hdvsend_param->flags & NO_AUDIO_STREAM)) {
    _add_dvdif_to_outbuf(hdvsend_param, difblock);
    if (hdvsend_param->flags & SEPARATED_AUDIO_STREAM) {
      _add_dvdif_to_audio_outbuf(hdvsend_param, difblock);
    }
  }

  if (hdvsend_param->audio_redundancy > 0) {
    _add_dvdif_to_r_outbuf(hdvsend_param, difblock);
  }
}

/* This will process DV VIDEO DIF blocks */
static void
_process_sct_video (struct hdvsend_param *hdvsend_param,
                    u_char seq, u_char dseq, u_char dbn, u_int32_t *difblock)
{
  if (hdvsend_param->flags & NO_VIDEO_STREAM) {
    /* do not include video data */
  } else if (hdvsend_param->frame_drop > 1 &&
      (frame % hdvsend_param->frame_drop) != 0) {
    /* drop this frame */
  } else {
    _add_dvdif_to_outbuf(hdvsend_param, difblock);
  }

  /* This DIF block is end of frame */
  if (hdvsend_param->format == DV_FORMAT_NTSC) {
    if (dseq == (DIFSEQ_NUM_NTSC - 1) && dbn == 134) {
      _dvdif_flush(hdvsend_param);
    }
  } else if (hdvsend_param->format == DV_FORMAT_PAL) {
    if (dseq == (DIFSEQ_NUM_PAL - 1) && dbn == 134) {
      _dvdif_flush(hdvsend_param);
    }
  }
}

static void
_add_dvdif_to_outbuf(struct hdvsend_param *hdvsend_param, u_int32_t *difblock)
{
  static int n;

  //printf("RTP sec[%d]->fecframe\n",rtphdr->seq);
  memcpy(&outbuf[(outbuf_len + sizeof(rtp_hdr_t))/4], difblock, 80);
  outbuf_len += 80;

  if (outbuf_len > 80*(hdvsend_param->DIFblocks_in_pkt - 1)) {
    n = send_pkt(hdvsend_param, outbuf, outbuf_len + sizeof(rtp_hdr_t));

    outbuf_len = 0;
    rtphdr->seq = htons(ntohs(rtphdr->seq)+1);
  }
  if(hdvsend_param->flags & USE_RATECONTROL){
	_dvrtp_sendto_ratecontrol(*hdvsend_param, NOTENDOFFRAME);
  }

  return;
}

static void
_add_dvdif_to_r_outbuf(struct hdvsend_param *hdvsend_param, u_int32_t *difblock)
{
  static int n;
  int i;

  memcpy(&r_outbuf[(r_outbuf_len + sizeof(rtp_hdr_t))/4], difblock, 80);
  r_outbuf_len += 80;

  if (r_outbuf_len > 80*(hdvsend_param->DIFblocks_in_pkt - 1)) {
    for (i=0; i<hdvsend_param->audio_redundancy; i++) {
      memcpy(r_outbuf, outbuf, sizeof(rtp_hdr_t));
      n = send_pkt(hdvsend_param, r_outbuf, r_outbuf_len + sizeof(rtp_hdr_t));
      rtphdr->seq = htons(ntohs(rtphdr->seq)+1);
    }

    r_outbuf_len = 0;
  }

  return;
}

static void
_add_dvdif_to_audio_outbuf(struct hdvsend_param *hdvsend_param,
                           u_int32_t *difblock)
{
  static int n;

  memcpy(&audiobuf[(audiobuf_len + sizeof(rtp_hdr_t))/4], difblock, 80);
  audiobuf_len += 80;

  if (audiobuf_len > 80*(hdvsend_param->DIFblocks_in_pkt - 1)) {
    n = send_audio_pkt(hdvsend_param,
                       audiobuf,
                       audiobuf_len + sizeof(rtp_hdr_t));

    if (n < 0) {
      perror("sendto audio");
    }

    audiobuf_len = 0;
    audio_rtphdr->seq = htons(ntohs(audio_rtphdr->seq)+1);
  }

  return;
}

static void
_dvdif_flush(struct hdvsend_param *hdvsend_param)
{
  if (outbuf_len > 0) {
    /* Send out the RTP packet.                         */
    /* In RTP, the data from different frame can not be */
    /* included in a single packet.                     */

    send_pkt(hdvsend_param, outbuf, outbuf_len + sizeof(rtp_hdr_t));
    outbuf_len = 0;
  }

  /* RTP sequence number will increase every time a RTP packet is sent */
  rtphdr->seq = htons(ntohs(rtphdr->seq)+1);

  /* RTP timestamp will increase every by frame.                     */
  /* since DV/RTP payload uses 90khz timestamp and NTSC is 29.97 hz, */
  /* the timestamp will simply increase by 3003.                     */
  if(hdvsend_param->format == DV_FORMAT_NTSC){
    rtphdr->ts = htonl(ntohl(rtphdr->ts)+DVNTSCRTPTSINC);
  } else if(hdvsend_param->format == DV_FORMAT_PAL){
    rtphdr->ts = htonl(ntohl(rtphdr->ts)+DVPALRTPTSINC);
  }
  /* The variable [frame] counts the number of frames read */
  /* from the IEEE 1394 device                             */
  frame++;

  /* audio will be sent to a different UDP port */
  /* when it needs to be sent separetly.        */
  if (hdvsend_param->flags & SEPARATED_AUDIO_STREAM) {
    if (audiobuf_len > 0) {
      send_audio_pkt(hdvsend_param, audiobuf, audiobuf_len + sizeof(rtp_hdr_t));
      audiobuf_len = 0;
      if(hdvsend_param->format == DV_FORMAT_NTSC){
	audio_rtphdr->ts = htonl(ntohl(audio_rtphdr->ts)+DVNTSCRTPTSINC);
      } else if(hdvsend_param->format == DV_FORMAT_PAL){
	audio_rtphdr->ts = htonl(ntohl(audio_rtphdr->ts)+DVPALRTPTSINC);
      }
      audio_rtphdr->seq = htons(ntohs(audio_rtphdr->seq)+1);
    }
  }
  if(hdvsend_param->flags & USE_RATECONTROL){
	_dvrtp_sendto_ratecontrol(*hdvsend_param, ENDOFFRAME);
  }

  if (hdvsend_param->frame_max != 0 &&
	frame > hdvsend_param->frame_max) {
	exit(EXIT_SUCCESS);
  }

  return;
}

struct timeval
timersub_nanosleep(long int tv_nsec){
        struct timeval tv_prev, tv_cur, tv_sub;
        struct timespec tsreq;
        memset( (void *)&tsreq, 0, sizeof(tsreq));
        tsreq.tv_nsec=tv_nsec;
        gettimeofday(&tv_prev, NULL);
        nanosleep(&tsreq, NULL);
        gettimeofday(&tv_cur, NULL);
        timersub(&tv_cur, &tv_prev, &tv_sub);
        return(tv_sub);
}

void
_dvrtp_sendto_ratecontrol(struct hdvsend_param hdvsend_param, int eofflag)
{
        struct timeval exec_prev;
        static struct timeval exec_cur;
        static struct timeval sleep_sub, exec_sub;
        static long int tv_nsec;
        int ipPktNumPerFrame = (hdvsend_param.format == DV_FORMAT_NTSC ?
                                (DV_SD_NTSC_FRAMESIZE / DV_DIF_SIZE) / hdvsend_param.DIFblocks_in_pkt :
                                (DV_SD_PAL_FRAMESIZE / DV_DIF_SIZE) / hdvsend_param.DIFblocks_in_pkt);
        static int loopcnt = 0;
        loopcnt++;
        if(!eofflag && (loopcnt % hdvsend_param.do_sleep_count == 0) ){
                if(loopcnt == hdvsend_param.do_sleep_count){
                        gettimeofday(&exec_prev, NULL);
                        memcpy(&exec_cur, &exec_prev, sizeof(struct timeval));
                        tv_nsec = hdvsend_param.dvrtp_interval * hdvsend_param.do_sleep_count;
#ifdef LINUX
                        if(tv_nsec > (1000000000 / HZ) * 2 &&
                           tv_nsec % (1000000000 / HZ) < 1000000000 / HZ){
                                tv_nsec -= 1000000000 / HZ;
                        }
#endif /* LINUX */
                } else {
                        tv_nsec += hdvsend_param.dvrtp_interval * hdvsend_param.do_sleep_count - (sleep_sub.tv_usec + exec_sub.tv_usec) * 1000;
                        if(tv_nsec > hdvsend_param.dvrtp_interval * hdvsend_param.do_sleep_count){
                                tv_nsec = hdvsend_param.dvrtp_interval * hdvsend_param.do_sleep_count;
                        }
                }
                if(tv_nsec > 0){
                        gettimeofday(&exec_prev, NULL);
                        if((hdvsend_param.do_sleep_count == 1 ||
                            ipPktNumPerFrame - loopcnt) > (sleep_sub.tv_usec * 1000 / tv_nsec) * hdvsend_param.do_sleep_count ||
                            hdvsend_param.do_sleep_count > ipPktNumPerFrame / 2 ){
                                sleep_sub = timersub_nanosleep(tv_nsec);
                        }
                        timersub(&exec_prev, &exec_cur, &exec_sub);
                        gettimeofday(&exec_cur, NULL);
                } else{
                        sleep_sub.tv_usec -= hdvsend_param.dvrtp_interval * hdvsend_param.do_sleep_count /1000;
                }
        } else if(eofflag){
                loopcnt = 0;
        }
}

static void
_m2ts_flush(struct hdvsend_param *hdvsend_param, u_int32_t next_sph)
{
  int n = 0;
  u_int32_t cycle_count;
  u_int32_t cycle_offset;

  n = send_pkt(hdvsend_param, outbuf, outbuf_len + sizeof(rtp_hdr_t));

  rtphdr->seq = htons(ntohs(rtphdr->seq) + 1);
  rtphdr->ts = htonl(ntohl(rtphdr->ts) + hdvsend_param->ieee1394io.time_sub);
  outbuf_len = 0;

  cycle_count = ((next_sph >> 12) & 0x1fff) % 8000;
  cycle_offset = next_sph & 0xfff;

  int32_t cycle_count_sub = 0;
  int32_t cycle_offset_sub = 0;

  if (cycle_offset < hdvsend_param->ieee1394io.prev_cycle_offset) {
    cycle_offset_sub = (3072 + cycle_offset) - hdvsend_param->ieee1394io.prev_cycle_offset;
    if (cycle_count > 0) { cycle_count--; }
    else { cycle_count= 7999; }
  } else
    cycle_offset_sub = cycle_offset - hdvsend_param->ieee1394io.prev_cycle_offset;
  
  if (cycle_count < hdvsend_param->ieee1394io.prev_cycle_count)
    cycle_count_sub = (8000 + cycle_count) - hdvsend_param->ieee1394io.prev_cycle_count;
  else
    cycle_count_sub = cycle_count - hdvsend_param->ieee1394io.prev_cycle_count;
  
  hdvsend_param->ieee1394io.time_sub = cycle_count_sub * 90000 / 8000 + cycle_offset_sub * 90000 / 24576000; 
  
  hdvsend_param->ieee1394io.prev_cycle_count = cycle_count;
  hdvsend_param->ieee1394io.prev_cycle_offset = cycle_offset;    

  return;
}

static void
_add_m2ts_to_outbuf(struct hdvsend_param *hdvsend_param, u_int32_t *m2tssp, size_t length)
{
  int i;
  int count = length / M2TSSPSIZE;
  
  if (outbuf_len > hdvsend_param->mtu - 48 - sizeof(rtp_hdr_t) - M2TSPSIZE * count) {
    _m2ts_flush(hdvsend_param, ntohl(m2tssp[0]));
  }
    
  for (i = 0; i < count; i++)
  {
    memcpy(&outbuf[(outbuf_len + sizeof(rtp_hdr_t))/4], m2tssp + i * 48 + 1, M2TSPSIZE);
    outbuf_len += M2TSPSIZE;
  }

  return;
}

static void
_add_m2ts_sp_to_outbuf(struct hdvsend_param *hdvsend_param, u_int32_t *m2tssp, size_t length)
{
  int count = length / M2TSSPSIZE;

  if (outbuf_len > hdvsend_param->mtu - 48 - sizeof(rtp_hdr_t) - M2TSSPSIZE * count) {
    _m2ts_flush(hdvsend_param, ntohl(m2tssp[0]));
  }
  
  memcpy(&outbuf[(outbuf_len + sizeof(rtp_hdr_t))/4], m2tssp, length);
  outbuf_len += length;

  return;
}
